//Copyright (C) 2004, Brian Enigma <enigma at netninja.com>
//This file is part of MagicCodes.
//
//MagicCodes is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//MagicCodes is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with Foobar; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package org.ninjasoft.magiccodes.logic;

import java.util.*;
import java.util.jar.*;
import java.io.*;

/**
 * This class will (slightly inefficiently) look for all classes that implement
 * a particular interface.  It is useful for plugins.  This is done by looking
 * through the contents of all jar files in the classpath, as well as performing
 * a recursive search for *.class files in the classpath directories.
 * 
 * This particular class may not work in restrictive ClassLoader environments
 * such as Applets or WebStart.  (It may...but unlikely and untested.) 
 * @author enigma
 */
public class PluginDiscoverer {
    /** Print debugging statements? */
    private static final boolean DEBUG = false;
    /** List of the folders in the classpath */
    private Vector classpathFolders = new Vector();
    /** List of the jars in the classpath */
    private Vector classpathJars = new Vector();
    
    /**
     * Construct the object and parse apart the classpath into directories
     * and jars.  The constructor simple parses and does not search because
     * the parsing is quick and easy.  The actual search work should NOT be
     * done in the constructor, but in the findMatchingPlugins class.
     */
    public PluginDiscoverer() {
        String classpath = System.getProperty("java.class.path");
        StringTokenizer st = new StringTokenizer(classpath, File.pathSeparator);
        while (st.hasMoreTokens()) {
            String item = st.nextToken();
            File f = new File(item);
            if (item.toLowerCase().endsWith(".jar") && f.isFile()) {
                classpathJars.add(item);
            } else if (f.isDirectory()) {
                classpathFolders.add(item);
            }
        }
    }

    /**
     * Given the class/interface we want to match and the name of a class
     * (which may be a filename such as "org/ninjasoft/Blah.class" or an 
     * actual class name such as "org.ninjasoft.Blah"), determine if the
     * class implements the interface.  
     * @param interfaceClass Class object representing the interface we are looking for
     * @param testClass Name of class or class file we are testing
     * @return corrected/normalized name of class if matches, otherwise null
     */
    private String checkIfClassMatches(Class interfaceClass, String testClass) {
        // Drop any trailing .class, if this is a filename
        if (testClass.toLowerCase().endsWith(".class"))
            testClass = testClass.substring(0, testClass.length() - 6);
        // Normalize slashes to dots, if this is a filename
        testClass = testClass.replace('\\', '.').replace('/', '.');
        // Drop any leading dots (for instance, if this was a filename that started with a slash)
        while (testClass.startsWith("."))
            testClass = testClass.substring(1);
        // If it is an internal class, forget about it
        if (testClass.indexOf('$') != -1)
            return null;
        if (DEBUG)
            System.out.println(" class:" + testClass);
        try{
            Class testClassObj = Class.forName(testClass);
            if (interfaceClass.isAssignableFrom(testClassObj)) {
                if (DEBUG)
                    System.out.println("MATCH!");
                return testClass;
            }
        }catch(ClassNotFoundException e) {
            //e.printStackTrace();
            System.out.println("Unable to load class " + testClass);
        }
        return null;
    }
    
    /**
     * Find all classes in the class path that implement the interface defined
     * in the Class object parameter.
     * @param c a Class object representing the interface we are looking for
     * @return an array of all the class names that implement that interface
     */
    public String[] findMatchingPlugins(Class c) {
        Vector results = new Vector();
        // Check the jar files
        for (Iterator i = classpathJars.iterator(); i.hasNext(); ) {
            String filename = (String) i.next();
            if (DEBUG)
                System.out.println("jar: " + filename);
            try{
                JarFile jar = new JarFile(filename);
                for (Enumeration item = jar.entries(); item.hasMoreElements(); ) {
                    JarEntry entry = (JarEntry) item.nextElement();
                    String name = entry.getName();
                    if (name.toLowerCase().endsWith(".class")) {
                        String classname = checkIfClassMatches(c, name);
                        if (classname != null)
                        	results.add(classname);
                    }
                }
            }catch(IOException e){
                //e.printStackTrace();
                System.out.println("Unable to open jar " + filename);
            }
        }
        // Check the classpath folders
        for (Iterator i = classpathFolders.iterator(); i.hasNext(); ) {
            String folder = (String) i.next();
            if (DEBUG)
                System.out.println("classfolder:" + folder);
            recursePath(c, results, folder, "");
        }
        // Flatten the output
        String[] resultArray = new String[results.size()];
        for (int i=0; i<results.size(); i++)
            resultArray[i] = (String) results.get(i);
        return resultArray;
    }
    
    /**
     * Recurse a path, looking for class files
     * @param c the interface we are looking for
     * @param results where to put matching classes
     * @param base base directory
     * @param path current location in the traversal
     */
    private void recursePath(Class c, Vector results, String base, String path) {
        File f = new File(base + File.separator + path);
        if (!f.isDirectory())
            return;
        File[] matches = f.listFiles(new classFilter());
        for (int i=0; i<matches.length; i++) {
            String classname = path + File.separator + matches[i].getName();
            classname = checkIfClassMatches(c, classname);
            if (classname != null)
                results.add(classname);
        }
        matches = f.listFiles(new directoryFilter());
        for (int i=0; i<matches.length; i++) {
            String folder = path + File.separator + matches[i].getName();
            recursePath(c, results, base, folder);
        }
    }
    
    /**
     * File.listFiles filter that only matches class files
     */
    public class classFilter implements FilenameFilter {
        public boolean accept(File dir, String name) {
            if (name.toLowerCase().endsWith(".class"))
                return true;
            return false;
        }
    }
    
    /**
     * File.listFiles filter that only matches subdirectories
     */
    public class directoryFilter implements FilenameFilter {
        public boolean accept(File dir, String name) {
            if (new File(dir.getPath() + File.separator + name).isDirectory())
                return true;
            return false;
        }
    }
    
    /**
     * Simple main function for testing
     * @param argv unused
     * @throws Exception if there was a problem
     */
    public static void main(String[] argv) throws Exception {
        PluginDiscoverer pd = new PluginDiscoverer();
        Class pluginInterface = Class.forName("org.ninjasoft.magiccodes.plugins.Plugin");
        String[] plugins = pd.findMatchingPlugins(pluginInterface);
        System.out.println();
        System.out.println("" + plugins.length + " plugin" + (plugins.length == 1 ? "" : "s") + " discovered" + (plugins.length == 0 ? "" : ":"));
        for (int i=0; i<plugins.length; i++)
            System.out.println(plugins[i]);
    }
}
